/*****************************************************************************\
 *
 * SoScreenPositioner.cpp
 *
 * Screen positioning node.
 * Acts as special camera and sets the orthographic projection 
 * in a special way convenient for menus and other stuff.
 * One common use is for the graphic items that shall stay 
 * on their places independently on the window size.
 *
 * Authors: PCJohn (peciva _at fit.vutbr.cz)
 * Contributors:
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel forced to do so.
 *
\*****************************************************************************/


#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoHandleEventAction.h>
#include <Inventor/actions/SoPickAction.h>
#include <Inventor/sensors/SoNodeSensor.h>
#include "SoScreenPositioner.h"



SO_NODE_SOURCE(SoScreenPositioner);


#define THIS this->pimpl



class SoScreenPositionerP {
public:
  SoScreenPositioner *parent;

  SbViewportRegion previousViewport;
  SoNodeSensor nodeSensor;
  static void nodeSensorCB(void *data, SoSensor *sensor);
  SbBool cameraNeedsUpdate;
  void updateCameraIfNecessary(const SbViewportRegion &vp);

  SoScreenPositionerP(SoScreenPositioner *parent)
  {
    this->parent = parent;
    nodeSensor.setFunction(nodeSensorCB);
    nodeSensor.setData(this);
    nodeSensor.setPriority(0); 
    nodeSensor.attach(parent); 
  }
};



void SoScreenPositioner::initClass()
{
  SO_NODE_INIT_CLASS(SoScreenPositioner, SoOrthographicCamera, "OrthographicCamera");
}


SoScreenPositioner::SoScreenPositioner()
{
  SO_NODE_CONSTRUCTOR(SoScreenPositioner);

  SO_NODE_ADD_FIELD(screenOrigin,     (0.5f, 0.5f));
  SO_NODE_ADD_FIELD(screenPosition,   (0,0));
  SO_NODE_ADD_FIELD(pixelsPerUnit,    (1));
  SO_NODE_ADD_FIELD(unitsPerScreen,   (0));

  pimpl = new SoScreenPositionerP(this);

  THIS->cameraNeedsUpdate = TRUE;
}


SoScreenPositioner::~SoScreenPositioner()
{
  delete pimpl;
}


void SoScreenPositioner::GLRender(SoGLRenderAction *action)
{
  THIS->updateCameraIfNecessary(action->getViewportRegion());
  inherited::GLRender(action);
}


void SoScreenPositioner::doAction(SoAction *action)
{
  inherited::doAction(action);
}


void SoScreenPositionerP::updateCameraIfNecessary(const SbViewportRegion &vp)
{
	
  if (!(vp == previousViewport) || cameraNeedsUpdate) {

    // window size
    SbVec2s winSize = vp.getWindowSize();

    // clip window dimensions to be >= 1
    if (winSize[0] <= 0)  winSize[0] = 1;
    if (winSize[1] <= 0)  winSize[1] = 1;

    // compute camera height value (in space units)
    float height = parent->unitsPerScreen.getValue();
    float ppu = parent->pixelsPerUnit.getValue();
    if (ppu!=0)  height += float(winSize[1]) / ppu;

    // compute camera width value (in space units)
    float width = height * float(winSize[0]) / winSize[1];

    // compute camera position based on screenOrigin and screenPosition field
    float x = (0.5f - parent->screenOrigin.getValue()[0]) * width;
    float y = (0.5f - parent->screenOrigin.getValue()[1]) * height;
    x -= parent->screenPosition.getValue()[0] * width  / float(winSize[0]);
    y -= parent->screenPosition.getValue()[1] * height / float(winSize[1]);

    // update camera values
    // note: Updating camera while renderAction is in progress will cause one
    // additional scene re-rendering after this rendering is finished.
    // This may be avoided by disabling notification, however it may have 
    // another bad consequences... 
    nodeSensor.detach();
    parent->height.setValue(height);
    parent->position.setValue(x, y, parent->position.getValue()[2]);
    nodeSensor.attach(parent);

    // update variables
    previousViewport = vp;
    cameraNeedsUpdate = FALSE;
  }
  
}


void SoScreenPositionerP::nodeSensorCB(void *data, SoSensor *sensor)
{
  SoScreenPositionerP *p = (SoScreenPositionerP*)data;
  if (((SoNodeSensor*)sensor)->getTriggerNode() != p->parent)
    return;

  SoField *f = ((SoNodeSensor*)sensor)->getTriggerField();

  // Camera settings may need to be changed
  if (f)  p->cameraNeedsUpdate = TRUE;
}


// This function is overrides SoOrthographicCamera::getViewVolume
// to avoid viewport adjusting that depends on window aspect ratio.
SbViewVolume SoScreenPositioner::getViewVolume(float useaspectratio) const
{
  SbViewVolume volume;

  switch (viewportMapping.getValue()) {
  case CROP_VIEWPORT_FILL_FRAME:
  case CROP_VIEWPORT_LINE_FRAME:
  case CROP_VIEWPORT_NO_FRAME: {
      // use original behaviour
      // Aspect ratio will be taken from SoCamera::aspectRatio and
      // view volume will be adjusted
      if (useaspectratio == 0.0f) useaspectratio = this->aspectRatio.getValue();
      float halfheight = this->height.getValue() * 0.5f;
      float halfwidth = halfheight * useaspectratio;
      volume.ortho(-halfwidth, halfwidth,
                   -halfheight, halfheight,
                   this->nearDistance.getValue(), this->farDistance.getValue());
      break;
  }

  case ADJUST_CAMERA: {
      // avoid any aspect ratio adjustments
      // Aspect ratio will be used from actual window size.
      float halfheight = this->height.getValue() * 0.5f;
      float halfwidth = halfheight * THIS->previousViewport.getViewportAspectRatio();
      volume.ortho(-halfwidth, halfwidth,
                   -halfheight, halfheight,
                   this->nearDistance.getValue(), this->farDistance.getValue());
      // eliminate effect of SoCamera::getView that scales viewport
      if (useaspectratio < 1.0f)
        volume.scale(useaspectratio);
      break;
  }

  case LEAVE_ALONE: {
      // avoid any aspect ratio adjustments
      // Aspect ratio will be used from actual window size as in the previous case.
      useaspectratio = THIS->previousViewport.getViewportAspectRatio();
      float halfheight = this->height.getValue() * 0.5f;
      float halfwidth = halfheight * useaspectratio;
      volume.ortho(-halfwidth, halfwidth,
                   -halfheight, halfheight,
                   this->nearDistance.getValue(), this->farDistance.getValue());
      break;
  }
  default:
      assert(0 && "unknown viewport mapping");
      break;
  };

  volume.rotateCamera(this->orientation.getValue());
  volume.translateCamera(this->position.getValue());
  return volume;
}


void SoScreenPositioner::getBoundingBox(SoGetBoundingBoxAction *action)
{
  THIS->updateCameraIfNecessary(action->getViewportRegion());
  inherited::getBoundingBox(action);
}


void SoScreenPositioner::pick(SoPickAction *action)
{
  // Note: The picking may behave strangely under some circumstancies.
  // To avoid them, place SoScreenPositioner related stuff at the end of your scene graph like:
  //
  //             root
  //               |
  //     ------------------------
  //    |     |     |            |
  //  Scene  .... ....       Separator
  //                             |
  //                        -----------------------------------------------------
  //                       |                   |                  |              |
  //                  ScreenPositioner  positionedGeometry  ScreenPositioner   .....
  //
  //
  // The reason is that there is known limitation of Open Inventor - there
  // is no camera element and SoSeparator can not restore camera changes made
  // by SoScreenPositioner. Although rendering may work correctly since camera
  // is restored by popping OpenGL matrix, this will not work through picking
  // usually resulting picking incorrect behaviour.
  THIS->updateCameraIfNecessary(action->getViewportRegion());
  inherited::pick(action);
}


void SoScreenPositioner::cleanup()
{
  SoScreenPositioner::atexit_cleanup();
}


void SoScreenPositioner::callback(SoCallbackAction *action)
{
  THIS->updateCameraIfNecessary(action->getViewportRegion());
  inherited::callback(action);
}


void SoScreenPositioner::getMatrix(SoGetMatrixAction *action)
{
  THIS->updateCameraIfNecessary(action->getViewportRegion());
  inherited::getMatrix(action);
}


void SoScreenPositioner::handleEvent(SoHandleEventAction *action)
{
  THIS->updateCameraIfNecessary(action->getViewportRegion());
  inherited::handleEvent(action);
}


void SoScreenPositioner::getPrimitiveCount(SoGetPrimitiveCountAction *action)
{
  inherited::getPrimitiveCount(action);
}


void SoScreenPositioner::audioRender(SoAudioRenderAction *action)
{
  inherited::audioRender(action);
}
